<?php

namespace Drupal\acquia_contenthub\EventSubscriber\PruneCdf;

use Acquia\ContentHubClient\CDF\CDFObject;
use Acquia\ContentHubClient\CDFDocument;
use Acquia\ContentHubClient\Settings;
use Drupal\acquia_contenthub\AcquiaContentHubEvents;
use Drupal\acquia_contenthub\Event\PruneCdfEntitiesEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Class HandleChannelLanguages.
 *
 */
class HandleChannelLanguages implements EventSubscriberInterface {

  /**
   * The client settings.
   *
   * @var \Acquia\ContentHubClient\Settings
   */
  protected $clientSettings;

  /**
   * HandleChannelLanguages constructor.
   *
   * @param \Acquia\ContentHubClient\Settings $client_settings
   *   The client settings
   */
  public function __construct(Settings $client_settings) {
    $this->clientSettings = $client_settings;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events[AcquiaContentHubEvents::PRUNE_CDF][] = 'onPruneCdf';
    return $events;
  }

  /**
   * Handles channel languages.
   *
   * @param \Drupal\acquia_contenthub\Event\PruneCdfEntitiesEvent $event
   *   The prune event.
   */
  public function onPruneCdf(PruneCdfEntitiesEvent $event) {
    $cdf = $event->getCdf();
    $removedLanguages = $this->removeNotChannelLanguages($cdf);

    if (empty($removedLanguages)) {
      return;
    }

    foreach ($cdf->getEntities() as $cdfEntity) {
      $this->removeDependenciesToNotChannelLanguages($cdfEntity, $removedLanguages);
    }
  }

  /**
   * Removes languages that don't belong to the channel.
   *
   * @param \Acquia\ContentHubClient\CDFDocument $cdf
   *   The CDF document.
   *
   * @return array
   *   The languages that don't belong to the channel.
   */
  protected function removeNotChannelLanguages(CDFDocument $cdf) {
    $name = $this->clientSettings->getName();

    $removedLanguages = [];
    foreach ($cdf->getEntities() as $cdfEntity) {
      $entity_type = $cdfEntity->getAttribute('entity_type')->getValue()['und'];

      if ($entity_type !== 'configurable_language') {
        continue;
      }

      // If language belongs to a channel and it isn't current channel, remove them from the CDF document.
      $channelNameAttribute = $cdfEntity->getAttribute('channel_name');
      if ($channelNameAttribute
        && !empty($channelName = $channelNameAttribute->getValue()['und'])
        && $channelName !== $name) {
        $removedLanguages[$cdfEntity->getUuid()] = $cdfEntity;
        $cdf->removeCdfEntity($cdfEntity->getUuid());
      }
    }

    return $removedLanguages;
  }

  /**
   * Removes languages that don't belong to the channel.
   *
   * @param \Acquia\ContentHubClient\CDF\CDFObject $cdfEntity
   *   The CDF object.
   * @param array $removedLanguages
   *   The languages that don't belong to the channel.
   */
  protected function removeDependenciesToNotChannelLanguages(CDFObject $cdfEntity, array $removedLanguages) {
    $metadata = $cdfEntity->getMetadata();

    // Not a multilingual entity, nothing to do here.
    if (!isset($metadata['languages'])) {
      return;
    }
    /**
     * @var CDFObject $languageCdf
     */
    foreach ($removedLanguages as $languageUuid => $languageCdf) {
      $removedLangcode = $languageCdf->getMetadata()['langcode'];
      if (($langcodeKey = array_search($removedLangcode, $metadata['languages'], TRUE)) !== FALSE) {
        unset($metadata['languages'][$langcodeKey]);
      }

      // If default language doesn't belong to channel, used the first language that does.
      if (!in_array($metadata['default_language'], $metadata['languages'], TRUE)) {
        $metadata['default_language'] = current($metadata['languages']);
      }

      unset($metadata['dependencies']['entity'][$languageUuid]);
    }
    $cdfEntity->setMetadata($metadata);
  }
}
